package com.bergerkiller.bukkit.nolagg.sending;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.LinkedList;
import java.util.logging.Level;
import java.util.Set;
import java.util.HashSet;

import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.block.BlockFace;
import org.bukkit.entity.Player;

import org.bukkit.craftbukkit.entity.CraftPlayer;

import com.all.bukkit.threadcheck.ThreadCheck;
import com.bergerkiller.bukkit.nolagg.NLListener;
import com.bergerkiller.bukkit.nolagg.NoLagg;
import com.bergerkiller.bukkit.nolagg.RemainderRate;
import com.bergerkiller.bukkit.nolagg.Util;

import net.minecraft.server.Chunk;
import net.minecraft.server.ChunkCoordIntPair;
import net.minecraft.server.EntityPlayer;
import net.minecraft.server.NetworkManager;
import net.minecraft.server.Packet;
import net.minecraft.server.Packet53BlockChange;
import net.minecraft.server.World;

@SuppressWarnings({"rawtypes", "unchecked"})
public class ChunkSendQueue extends LinkedList {

	private static ArrayList<Operation> syncedops;
	
	private static Field queuesizefield;
	public static boolean enabled = true;
	private static final long serialVersionUID = 1L;
	private static int task = -1;
	private static void executeSyncedOps() {
		synchronized (syncedops) {
			for (Operation op : syncedops) op.run();
			syncedops.clear();
		}
	}
	public static void init() {
		syncedops = new ArrayList<Operation>();
		try {
			queuesizefield = NetworkManager.class.getDeclaredField("x");
			queuesizefield.setAccessible(true);
		} catch (Throwable t) {
			NoLagg.log(Level.SEVERE, "Failed to hook into the player packet queue size field");
			NoLagg.log(Level.SEVERE, "Distortions in the chunk rate will cause players to get kicked");
		}
	}
	public static void updateAll() {
		try {
			//synced ops
			executeSyncedOps();
			if (enabled) {
				global(new Operation() {
					public void handle(EntityPlayer ep) {
						bind(ep).update();
					}
				});
			}
		} catch (Exception ex) {
			NoLagg.log(Level.SEVERE, "An error occured while sending chunks:");
			ex.printStackTrace();
		} catch (OutOfMemoryError ex) {
			NoLagg.log(Level.SEVERE, "We are running out of memory here!");
			NoLagg.log(Level.SEVERE, "Restart the server and increase the RAM usage available for Bukkit.");
		}
	}
	public static void deinit() {
		if (task != -1) {
			Bukkit.getServer().getScheduler().cancelTask(task);
		}
		executeSyncedOps();
		syncedops = null;
		//clear bound queues
		global(new Operation() {
			public void handle(EntityPlayer ep) {
				ChunkSendQueue queue = bind(ep);
				LinkedList list = new LinkedList(queue.contained);
				ep.chunkCoordIntPairQueue = list;
			}
		});
	}
	
	public static void scheduleQueueGlobal(final Chunk chunk) {
		scheduleQueueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void scheduleQueueGlobal(final World world, final int x, final int z) {
		synchronized(syncedops) {
			syncedops.add(new Operation() {
				public void init() {
					queueGlobal(world, x, z);
				}
			});
		}
	}
	
	public static void deQueueGlobal(final Chunk chunk) {
		deQueueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void queueGlobal(final Chunk chunk) {
		queueGlobal(chunk.world, chunk.x, chunk.z);
	}
	public static void deQueueGlobal(final World world, final int x, final int z) {
		deQueueGlobal(world, new ChunkCoordIntPair(x, z));
	}
	public static void queueGlobal(final World world, final int x, final int z) {
		queueGlobal(world, new ChunkCoordIntPair(x, z));
	}
	public static void deQueueGlobal(final World world, final ChunkCoordIntPair pair) {
		if (!ThreadCheck.check(null)) return;
		world(world, new Operation() {
			public void handle(EntityPlayer ep) {
				ep.chunkCoordIntPairQueue.remove(pair);
			}
		});
	}
	public static void queueGlobal(final World world, final ChunkCoordIntPair pair) {
		if (!ThreadCheck.check(null)) return;
		world(world, new Operation() {
			public void handle(EntityPlayer ep) {
		        if (Math.abs(pair.x - Util.toChunk(ep.locX)) > Util.view) return;
		        if (Math.abs(pair.z - Util.toChunk(ep.locZ)) > Util.view) return;
				ep.chunkCoordIntPairQueue.add(0, pair);
			}
		});
	}
	
	private int maxQueueSize = 300000;
	private int packetBufferQueueSize = 0;
	private int updateQueueSize() {
		if (queuesizefield != null) {
			try {
				this.packetBufferQueueSize = (Integer) queuesizefield.get(this.ep.netServerHandler.networkManager);
				while (this.packetBufferQueueSize < 0) {
					this.packetBufferQueueSize += 9437184;
				}
			} catch (Throwable t) {
				queuesizefield = null;
				NoLagg.log(Level.SEVERE, "Failed to hook into the player packet queue size field");
				NoLagg.log(Level.SEVERE, "Distortions in the chunk rate will cause players to get kicked");
			}
		}
		return this.packetBufferQueueSize;
	}
	private int getMaxQueueSize() {
		return NoLagg.isSpoutEnabled ? 10485760 : 1048576;
	}
	
	private static <T extends Operation> T global(T op) {
		op.execute();
		return op;
	}
	private static <T extends Operation> T world(World world, T op) {
		op.execute(world);
		return op;
	}
	private static class Operation {
		public void handle(EntityPlayer ep) {}
		public void init(World world) {}
		public void init() {}
		public void execute(World world) {
			ThreadCheck.check(null);
			this.init(world);
			try {
				for (Object o : world.players) this.handle((EntityPlayer) o);
			} catch (ConcurrentModificationException ex) {
				NoLagg.log(Level.SEVERE, "A plugin tried to modify 'World.players' (world entity list) from an async thread!");
			}
		}
		public void execute() {
			ThreadCheck.check(null);
			this.init();
			for (World world : Util.getWorlds()) {
				this.execute(world);
			}
		}
		public Operation run() {
			this.execute();
			return this;
		}
	}
	
	public static ChunkSendQueue bind(Player with) {
		if (!enabled) return null;
		return bind(((CraftPlayer) with).getHandle());
	}
	public static ChunkSendQueue bind(EntityPlayer with) {
		if (!enabled) return null;
		if (!(with.chunkCoordIntPairQueue instanceof ChunkSendQueue)) {
			ThreadCheck.check(null);
			with.chunkCoordIntPairQueue = new ChunkSendQueue(with);
		}
		return (ChunkSendQueue) with.chunkCoordIntPairQueue;
	}	
	private ChunkSendQueue(EntityPlayer ep) {
		this.ep = ep;
		this.addAll(ep.chunkCoordIntPairQueue);
		this.world = ep.world;
		this.sendDirection = Util.getDirection((int) this.ep.yaw - 90);
		this.x = (int) (ep.locX + ep.motX * 16) >> 4;
		this.z = (int) (ep.locZ + ep.motZ * 16) >> 4;
		this.chunkQueue = new ChunkCompressQueue(this);
		ChunkCompressionThread.addQueue(this.chunkQueue);
	}
		
	public static double maxRate = 2;
	public static double minRate = 0.25;
			
	public static double getAverageRate() {
		return global(new Operation() {
			double totalrate = 0;
			int pcount = 0;
			public void handle(EntityPlayer ep) {
				totalrate += bind(ep).rate.get();
				pcount++;
			}
			public double get() {
				return this.totalrate / (double) pcount;
			}
		}).get();
	}
	public double getRate() {
		return this.rate.get();
	}
	
	public EntityPlayer ep;
	public BlockFace sendDirection = BlockFace.NORTH;
	public World world;
	public int x;
	public int z;
	private RemainderRate rate = new RemainderRate(2);
	private ChunkCompressQueue chunkQueue;
	private final LinkedList<Packet53BlockChange> toTrigger = new LinkedList<Packet53BlockChange>();
	private final Set<ChunkCoordIntPair> contained = new HashSet<ChunkCoordIntPair>();
	public void removeContained(int x, int z) {
	    this.removeContained(new ChunkCoordIntPair(x, z));
	}
	public void removeContained(ChunkCoordIntPair pair) {
		this.contained.remove(pair);
	}
	
	public void scheduleTrigger(Chunk chunk) {
    	this.scheduleTrigger(new Packet53BlockChange(chunk.x << 4, 0, chunk.x << 4, chunk.world));
	}
	public void scheduleTrigger(Packet53BlockChange packet) {
		this.toTrigger.offer(packet);
	}
	public boolean sendNextTrigger() {
		Packet53BlockChange p = this.toTrigger.poll();
		if (p == null) return false;
		this.send(p);
		return true;
	}
	
	/*
	 * Used to send packets and triggers
	 */
	private int intervalcounter = -1;
	private int buffersizeavg = 0;
	public String getBufferLoadMsg() {
		double per = Util.round(100D * this.buffersizeavg / getMaxQueueSize(), 2);
		if (this.buffersizeavg > 300000) {
			return ChatColor.RED.toString() + per + "%";
		} else if (this.buffersizeavg > 100000) {
			return ChatColor.GOLD.toString() + per + "%";
		} else {
			return ChatColor.GREEN.toString() + per + "%";
		}
	}
	
	public void send(Packet packet) {
		this.ep.netServerHandler.networkManager.queue(packet);
	}
	
	private int prevQueueSize = 0;
	private void update() {
		this.updateQueueSize();
		if (this.buffersizeavg == 0) {
			this.buffersizeavg = this.packetBufferQueueSize;
		} else {
			this.buffersizeavg += 0.3 * (this.packetBufferQueueSize - this.buffersizeavg);
		}
		if (super.isEmpty() && !this.chunkQueue.canSend() && this.toTrigger.isEmpty()) return;
		double newrate = this.rate.get();
		if (this.packetBufferQueueSize > this.maxQueueSize) {
			newrate = minRate;
		} else {
			if (this.prevQueueSize > this.packetBufferQueueSize) {
				newrate += 0.08;
			} else {
				//to force the rate to be optimal
				if (this.packetBufferQueueSize > 80000) {
					newrate -= 0.1;
				} else if (this.packetBufferQueueSize > 20000) {
					newrate -= 0.05;
				} else {
					newrate += 0.05;
				}
			}
			newrate += 0.9 * (this.rate.get() - newrate);
			//set rate bounds
			if (newrate > maxRate) {
				newrate= maxRate;
			} else if (newrate < minRate) {
				newrate = minRate;
			}
		}

		
		this.rate.set(newrate);
		this.prevQueueSize = this.packetBufferQueueSize;
		//send chunks
		if (newrate >= 1) {
			this.update(1, (int) this.rate.next());
		} else {
			this.update((int) (1 / this.rate.get()), 1);
		}
	}
	public Object[] toArray(Object[] value) {
		synchronized (this) {
			return super.toArray(value);
		}
	}
	
	private class IndexedPair extends ChunkCoordIntPair implements Comparable<IndexedPair> {
		public int index;
		public IndexedPair(final int x, final int z) {
			super(x, z);
		}
		public void calculateIndex(final int midx, final int midz, final BlockFace direction) {
			this.index = ChunkCoordIndexer.get(this.x - midx, this.z - midz, direction);
		}
		public int compareTo(IndexedPair ipair) {
			return this.index - ipair.index;
		}
	}
	
	private void update(int interval, int rate) {
		if (interval == 0) interval = 1;
		if (rate == 0) rate = 1;
		if (this.intervalcounter >= interval)  {
			//sorting
			BlockFace newDirection = Util.getDirection((int) this.ep.yaw - 90);
			int newx = (int) (ep.locX + ep.motX * 16) >> 4;
			int newz = (int) (ep.locZ + ep.motZ * 16) >> 4;
			if (ep.world != this.world || newx != this.x || newz != this.z || this.sendDirection != newDirection) {
				this.sendDirection = newDirection;
				this.x = newx;
				this.z = newz;
				this.world = ep.world;
				this.chunkQueue.update(this.world, this.x, this.z, this.sendDirection);
				//sort self
				synchronized (this) {
					if (!super.isEmpty()) {
						this.doContainCheck = false;
						try {
							//Calculate indices
							for (Object o : this) ((IndexedPair) o).calculateIndex(this.x, this.z, this.sendDirection);
							java.util.Collections.sort(this);
						} catch (ConcurrentModificationException ex) {
							NoLagg.log(Level.SEVERE, "A plugin tried to modify 'EntityPlayer.chunkCoordIntPairQueue' from another thread!");
						} catch (ArrayIndexOutOfBoundsException ex) {
							NoLagg.log(Level.SEVERE, "A plugin tried to modify 'EntityPlayer.chunkCoordIntPairQueue' from another thread!");
						}
						this.doContainCheck = true;
					}
				}
			}
			this.sendBatch(rate);
			this.intervalcounter = 1;
		} else {
			this.intervalcounter++;
		}
	}
	
	private ChunkCoordIntPair pollPair() {
		while (!super.isEmpty()) {
			ChunkCoordIntPair pair = (ChunkCoordIntPair) super.poll();
			if (pair == null) return null;
			if (this.isNear(pair, Util.view)) {
				return pair;
			} else {
				this.contained.remove(pair);
			}
		}
		return null;
	}
	
	private void sendBatch(int count) {
		//send triggers
		for (int i = 0; i < count && this.sendNextTrigger(); i++);
				
		//load chunks
		for (int i = 0; i < count; i++) {
			ChunkCoordIntPair pair = this.pollPair();
			if (pair == null) break;
			NLListener.wasChunkLightingFixed = false;
			Chunk chunk = this.ep.world.getChunkAt(pair.x, pair.z);
			if (!NLListener.wasChunkLightingFixed) {
				//lighting fixer is currently not fixing this chunk
				//enqueuing is allowed
				this.chunkQueue.enqueue(chunk);
			} else {
				this.contained.remove(pair);
			}
		}
						
		//send chunks
		for (int i = 0; i < count && this.chunkQueue.sendNext(); i++);
	}
			
	public boolean containsAll(Collection coll) {
		synchronized (this) {
			return super.containsAll(coll);
		}
	}
	public boolean contains(Object o) {
		synchronized (this) {
			return super.contains(o);
		}
	}
	public boolean addAll(Collection coll) {
		synchronized (this) {
			for (Object o : coll) this.add(o);
			return true;
		}
	}
	public boolean addAll(int index, Collection coll) {
		synchronized (this) {
			for (Object o : coll) this.add(index, o);
			return true;
		}
	}
	public void clear() {
		synchronized (this) {
			super.clear();
		}
		this.chunkQueue.clear();
		this.contained.clear();
	}
	
	/*
	 * Prevent Spout from using this queue...seriously!
	 */
	public boolean isEmpty() {
		return true;
	}
	
	public int getPendingSize() {
		return this.size() + this.chunkQueue.getPendingSize();
	}
	
	/*
	 * get(0) is called in EntityPlayer to get the next chunk to send
	 */
	public Object get(int index) {
		return null;
	}
	public boolean remove(Object object) {
		ChunkCoordIntPair pair = (ChunkCoordIntPair) object;
		synchronized (this) {
			if (super.remove(object)) {
				this.contained.remove(pair);
				return true;
			}
		}
		if (this.chunkQueue.remove(pair.x, pair.z)) {
			this.contained.remove(pair);
			return true;
		}
		return false;
	}
	
	public boolean isNear(ChunkCoordIntPair coord, final int view) {
		return this.isNear(coord.x, coord.z, view);
	}
	public boolean isNear(final int chunkx, final int chunkz, final int view) {
        if (Math.abs(chunkx - Util.toChunk(this.ep.locX)) > view + 1) return false;
        if (Math.abs(chunkz - Util.toChunk(this.ep.locZ)) > view + 1) return false;
        return true;
	}
	
	private boolean doContainCheck = true;
	
	/*
	 * add(o) is called in PlayerInstance to queue a new chunk coordinate
	 */
	public void add(int index, Object object) {
		synchronized (this) {
			if (object == null) return;
			ChunkCoordIntPair pair = (ChunkCoordIntPair) object;
			if (!this.isNear(pair, Util.view)) return;
			if (doContainCheck && !this.contained.add(pair)) return;
			this.sendDirection = null; //invalidate
			super.add(index, new IndexedPair(pair.x, pair.z));
		}
	}
	public boolean add(Object object) {
		synchronized (this) {
			if (object == null) return false;
			ChunkCoordIntPair pair = (ChunkCoordIntPair) object;
			if (!this.isNear(pair, Util.view)) return false;
			if (doContainCheck && !this.contained.add(pair)) return false;
			this.sendDirection = null; //invalidate
			return super.add(new IndexedPair(pair.x, pair.z));
		}
	}

}
